from __future__ import annotations
from typing import Callable, List
import json
from hmac import HMAC
from hashlib import sha256

from trezorlib.misc import encrypt_keyvalue
from trezorlib.tools import parse_path
from trezorlib.client import TrezorClient
from trezorlib.exceptions import TrezorException, TrezorFailure
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from trezorpass.utils import prompt_trezor

from ..crypto import PATH, FILENAME_MESS, decrypt
from ..entry import Entry
from .managers import Manager, DropboxManager, FileManager
from .store_location import StoreLocation
from ..tag import Tag

class Store:
    def __init__(self, client: TrezorClient, manager_factory: Callable[[Store], Manager]):
        self.client = client
        self.entries: List[Entry] = []
        self.tags: List[Tag] = []
        self._lazy_master_key = None
        self._manager = manager_factory(self)

    @property
    def _master_key(self) -> str:
        if not self._lazy_master_key:
            self._lazy_master_key = self._get_master_key(self.client)
        return self._lazy_master_key

    @staticmethod
    def _get_master_key(client: TrezorClient) -> str:
        address_n = parse_path(PATH)
        key = "Activate TREZOR Password Manager?"
        value = bytes.fromhex("2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee")
        prompt_trezor()
        try:
            return encrypt_keyvalue(client, address_n, key, value).hex()
        except TrezorException:
            raise
        except:
            raise TrezorFailure()

    @property
    def filename(self) -> str:
        file_key = self._master_key[:len(self._master_key) // 2]
        return HMAC(file_key.encode(), FILENAME_MESS.encode(), sha256).hexdigest() + '.pswd'

    @staticmethod
    def _decrypt_store(encryptedStore: bytes, store: Store) -> str:
        '''Returns JSON representation of the password store'''
        enc_key = store._master_key[len(store._master_key) // 2:]
        try:
            return decrypt(enc_key, encryptedStore).decode('utf8')
        except:
            raise StoreDecryptError()

    @staticmethod
    def _decode_store(json_store: str, store: Store) -> Store:
        '''Returns object representation of the password store'''
        try:
            dict = json.loads(json_store)
            tags = {int(key): Tag.load(dict['tags'][key]) for key in dict['tags']}
            store.tags = [tags[key] for key in tags]
            store.entries = [Entry.load(dict['entries'][key], tags) for key in dict['entries']]
            return store
        except:
            raise StoreDecodeError()

    @staticmethod
    async def load(client: TrezorClient) -> Store:
        location = await inquirer.select("Where is the password store located at?", choices=[
            Choice(StoreLocation.Dropbox, "Dropbox"),
            Choice(StoreLocation.Filepath, "Filepath"),
        ]).execute_async()
        if location == StoreLocation.Dropbox:
            store = Store(client, lambda store: DropboxManager(store.filename))
        elif location == StoreLocation.Filepath:
            store = Store(client, lambda store: FileManager())
        else:
            raise Exception("Unreachable code")

        json_store = Store._decrypt_store(await store._manager.password_store, store)
        return Store._decode_store(json_store, store)

class StoreDecryptError(Exception):
    pass

class StoreDecodeError(Exception):
    pass


import os

from InquirerPy.utils import color_print

PRIMARY_COLOR = "#00783d"
SECONDARY_COLOR = "#f7931a"

os.environ["INQUIRERPY_STYLE_QUESTIONMARK"] = PRIMARY_COLOR
os.environ["INQUIRERPY_STYLE_ANSWERMARK"] = PRIMARY_COLOR
os.environ["INQUIRERPY_STYLE_ANSWER"] = SECONDARY_COLOR
os.environ["INQUIRERPY_STYLE_INPUT"] = SECONDARY_COLOR
os.environ["INQUIRERPY_STYLE_POINTER"] = SECONDARY_COLOR
os.environ["INQUIRERPY_STYLE_FUZZY_PROMPT"] = PRIMARY_COLOR

def prompt_trezor():
    print("Proceed on your trezor device")

def welcome():
    head = """####################################################
#                                                  #
#       #######                                    #
#     ##       ##                                  #
#    ##         ##                                 #
#    ##         ##                                 #
#  #################                               #
#  ##             ##                               #
#  ##             ##                               #
#  ##             ##  """
    text = "Trezor Password Manager CLI"
    tail = """  #
#  ##             ##                               #
#  ##             ##                               #
#  ###           ###                               #
#    ###       ###                                 #
#      #### ####                                   #
#         ###                                      #
#                                                  #
####################################################"""
    color_print([("class:banner", head), ("class:text", text), ("class:banner", tail)], {"banner": PRIMARY_COLOR, "text": SECONDARY_COLOR})

def goodbye():
    color_print([("class:padding", "##################### "), ("class:title", "Goodbye"),("class:padding", " ######################")], {"padding": PRIMARY_COLOR,"title": SECONDARY_COLOR})

def animate_dots(max_dots: int):
    dots = 0
    while True:
        yield
        if dots == max_dots:
            print("\b \b" * dots, end="", flush=True)
            dots = 0
        else:
            print(".", end="", flush=True)
            dots += 1


from typing import List
import time
import asyncio

from trezorlib.client import get_default_client, TrezorClient
from trezorlib.transport import TransportException
from trezorlib.exceptions import Cancelled, PinException, TrezorFailure

from InquirerPy import inquirer
from InquirerPy.utils import patched_print

from pyperclip import copy

from trezorpass.entry import Entry
from trezorpass.store import Store
from trezorpass.store.store import StoreDecodeError, StoreDecryptError
from trezorpass.utils import animate_dots, welcome, goodbye

async def select_entry(entries: List[Entry], long_instruction: str) -> Entry:
    '''Selects an entry'''
    choices = [{"value": entry, "name":entry.note if entry.note is not None else entry.title} for entry in entries]
    selection = await inquirer.fuzzy(
        message="Select an entry:",
        choices=choices,
        long_instruction=long_instruction
    ).execute_async()
    return selection

async def get_client() -> TrezorClient:
    '''Gets the Trezor client'''
    print("Looking for a Trezor device ", end="", flush=True)
    animation = animate_dots(5)
    while True:
        try:
            client = get_default_client()
            asyncio.get_event_loop().create_task(client_healthcheck(client))
            print()
            print("Trezor device ready")
            return client
        except TransportException:
            next(animation)
        except KeyboardInterrupt:
            print()
            raise
        except:
            print()
            print("Unable to access a Trezor device")
            raise
        await asyncio.sleep(0.7)

async def load_store(client: TrezorClient) -> Store:
    while True:
        try:
            return await Store.load(client)
        except StoreDecryptError:
            print("Unable to decrypt the password store")
        except StoreDecodeError:
            print("Unable to decode the password store")
        except (KeyboardInterrupt, asyncio.CancelledError):
            raise
        except:
            print("Unable to load the password store")

async def manage_entry(entry: Entry, client: TrezorClient) -> None:
    clipboard_dirty = False
    try:
        long_instruction = "Press Ctrl+C to leave the entry"
        while True:
            action = await inquirer.select("Select an action:", ['Show entry', 'Copy to clipboard', 'Show entry including secrets'], long_instruction=long_instruction).execute_async()
            if action == "Show entry":
                print(entry.show(client))
            elif action == "Copy to clipboard":
                key = await inquirer.select("Select the value to copy:", ['username', 'password'], long_instruction=long_instruction).execute_async()
                if key == 'username':
                    copy(entry.username)
                elif key == 'password':
                    clipboard_dirty = True
                    copy(entry.password_cleartext(client))
                print(f"{key} has been copied to the clipboard")
            elif action == "Show entry including secrets":
                print(entry.show(client, secrets=True))
    except KeyboardInterrupt:
        pass
    finally:
        if clipboard_dirty:
            print("Cleaning up the clipboard")
            copy("")

async def cli():
    welcome()
    try:
        client = await get_client()
        store = await load_store(client)
        while True:
            try:
                entry = await select_entry(store.entries, long_instruction="Press Ctrl+C to exit")
                await manage_entry(entry, client)
            except Cancelled:
                print("Trezor operation has been cancelled")
            except PinException:
                print("Invalid pin")
    except KeyboardInterrupt:
        pass
    except asyncio.CancelledError as e:
        message = e.args[0] if len(e.args) > 0 else "Unexpected cancellation"
        print(message)
    except:
        print("Unexpected error")
        exit(1)
    finally:
        if client:
            client.end_session()
        goodbye()

async def client_healthcheck(client: TrezorClient):
    while True:
        try:
            await asyncio.get_event_loop().run_in_executor(None, client.ping, "healthcheck")
        except:
            for task in asyncio.tasks.all_tasks():
                if not task.cancelled():
                    task.cancel("Trezor device has been disconnected")
            return
        await asyncio.sleep(1)

def run_cli():
    asyncio.run(cli())

if __name__ == "__main__":
    run_cli()


class Tag:
    def __init__(self, title: str) -> None:
        self.title = title

    @staticmethod
    def load(dict):
        tag = Tag(dict["title"])
        return tag

    def __str__(self) -> str:
        return self.title


import os
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

PATH = "m/10016'/0"
FILENAME_MESS = '5f91add3fa1c3c76e90c90a3bd0999e2bd7833d06a483fe884ee60397aca277a'

if os.getenv("PASSPHRASE") is None:
    os.environ["PASSPHRASE"] = ""

CIPHER_IVSIZE = 96 // 8;
AUTH_SIZE = 128 // 8;

def decrypt(key: str, data: bytes) -> bytes:
    iv = data[:CIPHER_IVSIZE]
    auth_tag = data[CIPHER_IVSIZE: CIPHER_IVSIZE + AUTH_SIZE]
    ciphertext = data[CIPHER_IVSIZE + AUTH_SIZE:]
    cipher = Cipher(algorithms.AES(bytes.fromhex(key)), modes.GCM(iv, auth_tag))
    decryptor = cipher.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()


from datetime import datetime
from pathlib import Path
import json
import os

import dropbox
import appdirs
from InquirerPy import inquirer

from trezorpass.store.managers.manager import Manager

DROPBOX_APP_KEY = "s340kh3l0vla1nv" # APP_KEY of the official TPM, potentionally breaking if maintainers disable PKCE flow for Dropbox auth
APP_DIR = appdirs.user_data_dir('trezor-pass')
DROPBOX_TOKEN_FILE = os.path.join(APP_DIR, 'dropbox')

class DropboxManager(Manager):
    def __init__(self, filename: str):
        self.filename = filename
        try:
            with open(DROPBOX_TOKEN_FILE, 'r') as file:
                self.oauth = json.load(file)
        except:
            self.oauth = None

    async def authenticate(self):
        oauth_result = None
        while not oauth_result:
            auth_flow = dropbox.DropboxOAuth2FlowNoRedirect(DROPBOX_APP_KEY, use_pkce=True, token_access_type='offline')
            authorize_url = auth_flow.start()
            print("1. Go to: " + authorize_url)
            print("2. Click \"Allow\" (you might have to log in first).")
            print("3. Copy the authorization code.")
            auth_code = await inquirer.text("Enter the authorization code here:").execute_async()
            try:
                oauth_result = auth_flow.finish(auth_code)
            except:
                retry = await inquirer.confirm("Invalid authorization code, retry?").execute_async()
                if not retry:
                    raise Exception()
        self.oauth = {
            "access_token": oauth_result.access_token,
            "refresh_token": oauth_result.refresh_token,
            "expiration": oauth_result.expires_at.isoformat()
            }
        try:
            Path(APP_DIR).mkdir(parents=True, exist_ok=True)
            with open(DROPBOX_TOKEN_FILE, 'w+') as file:
                os.chmod(DROPBOX_TOKEN_FILE, 0o600)
                file.write(json.dumps(self.oauth))
        except Exception as ex:
            pass

    @property
    async def password_store(self) -> bytes:
        if not self.oauth or not await inquirer.confirm("Credentials found, proceed?", default=True).execute_async():
            await self.authenticate()

        with dropbox.Dropbox(
            oauth2_access_token=self.oauth["access_token"],
            oauth2_refresh_token=self.oauth["refresh_token"],
            oauth2_access_token_expiration=datetime.fromisoformat(self.oauth["expiration"]),
            app_key=DROPBOX_APP_KEY
        ) as dbx:
            (_, response) = dbx.files_download("/" + self.filename)
            return response.content